""" Visualization Modules for NURBS-Python

.. moduleauthor:: Onur Rauf Bingol <orbingol@gmail.com>

"""

__author__ = "Onur Rauf Bingol"
__version__ = "2.0.0"
__license__ = "MIT"

# Import core library visualization module
from .. import linalg
from .. import vis
